(:~
 : This module handles the creation of all necessary files for displaying the
 : index information for each BdN author.
 :
 : The encoded indices are:
 : * an index of Bible passages
 : * an index of mentioned persons
 : * an index of mentioned Classics
 : * an index of mentioned subjects
 :
 : To improve performance this modules creates a separate XML file for each type
 : of index and each author in resources/aux. The entries in the files are
 : already sorted alphabetically resp. according to the Loccum norms for the
 : Bible.
 :
 : Furthermore the basic HTML files for an index are created; this encompasses
 : mainly the list view.
 :
 : @author Michelle Weidling
 : @version 1.1
 :
  :)

xquery version "3.1";

module namespace index-creation="http://bdn-edition.de/ns/index-creation";

declare namespace bdn="http://www.bdn-edition.de/ns/1.0";
declare namespace map = "http://www.w3.org/2005/xpath-functions/map";
declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace xhtml="http://www.w3.org/1999/xhtml";
declare namespace html="http://www.w3.org/1999/html";
declare namespace xi="http://www.w3.org/2001/XInclude";
declare namespace templates="http://exist-db.org/xquery/templates";
declare namespace textgrid="http://textgrid.info/namespaces/metadata/core/2010";

import module namespace bdn-utils="http://bdn-edition.de/ns/utils" at "utils.xqm";
import module namespace config="https://sade.textgrid.de/ns/config" at "../config.xqm";
import module namespace functx="http://www.functx.com";
import module namespace whitelist="https://bdn-edition.de/ns/whitelist" at "whitelist.xqm";

declare variable $index-creation:bible-map :=
  map:merge
  (
    for $book in doc($config:sade-data || "2r573.xml")//tei:bibl[@ana] return
      map:entry($book/@ana/string(), count($book/preceding-sibling::tei:bibl[@ana]) + 1)
  );

declare variable $index-creation:id-file-map :=
  map:merge
  (
    for $id in doc($config:aux || "id-file.xml")//bdn:id return
      map:entry($id/string(), $id/../@uri/string())
  );

declare variable $index-creation:index-types := ("bible-index", "persons-index",
  "subjects-index", "classics-index");

declare variable $index-creation:whitelist := doc($config:aux || "whitelist.xml");

(:~
 : Creates XML files for all sorts of indices and all authors for whom a full
 : XML file is available.
 :
 : TODO: This function should be called by post-install.xql
 :
 : @author Michelle Weidling
 : @return xs:string the locations of the stored files
 :)
declare function index-creation:main() as xs:string*
{
  let $full-xmls := collection($config:full-xmls)[matches(base-uri(.), "full.xml")]
  for $xml in $full-xmls return
    let $author := functx:substring-after-last(base-uri($xml), "/") => substring-before("-")
    let $determine-indices := index-creation:determine-occurring-indices($xml, $author)
    let $author-node := $index-creation:whitelist//bdn:author[./string() = $author]
    let $occuring-indices := $author-node/following-sibling::bdn:index/@type/string()

    let $xml :=
      for $index in $occuring-indices return
        switch ($index)
          case "bible-index" return
            index-creation:make-bible-index-xml($author, $xml)
          default return
            index-creation:make-index-xml($author, $xml, $index)
      let $html := index-creation:make-all-htmls($author, $occuring-indices)
      return
          "Successfully created all indices incl. HTML serialization."
};

(:~
 : Creates XML files for all sorts of indices and a specific author for whom a full
 : XML file is available.
 :
 : @author Michelle Weidling
 : @param $author The name of the author
 : @return xs:string the locations of the stored files
 :)
declare function index-creation:index-for-author($author as xs:string, $log as xs:string)
as xs:string
{
    let $add-log-entry := bdn-utils:add-log-entry($log, "Start index creation")

    let $xml :=
        if (collection($config:full-xmls)[matches(base-uri(.), $author)]) then
            collection($config:full-xmls)[matches(base-uri(.), $author)]
        else
            error( QName("http://bdn-edition.de/ns/index-creation", "INDEX001"), "No full XML file found for " || $author)


    let $index-map := index-creation:determine-occurring-indices($xml, $author)
    let $occuring-indices :=
        function($index-type, $occurs)
        {
            if ($occurs) then
                $index-type
            else
                ()
        }
    let $occuring-indices := map:for-each($index-map, $occuring-indices)

    let $xml :=
      for $index in $occuring-indices return
        (bdn-utils:add-log-entry($log, "Start creation of " || $index),
        switch ($index)
          case "bible-index" return
            index-creation:make-bible-index-xml($author, $xml)
          default return
            index-creation:make-index-xml($author, $xml, $index))
      let $html := index-creation:make-all-htmls($author, $occuring-indices)
      return
          "Successfully created all indices incl. HTML serialization."
};


(:~
 : Creates an author specific XML file for the Bible index.
 :
 : @author Michelle Weidling
 : @param $author The current author
 : @param $xml The current and opened full XML file (e.g. griesbach-full.xml)
 : @return xs:string the locations of the stored files
 :)
declare function index-creation:make-bible-index-xml($author as xs:string,
$xml as node()*) as xs:string*
{
  let $entries := $xml//tei:bibl[@type = "biblical-reference"]/tei:citedRange
  let $extended-entries :=
    for $entry in $entries
      let $chapterID := $entry/ancestor::tei:div[@xml:id][1]/@xml:id
      let $key :=
        if ($entry//tei:lem) then
            (: textual criticism can be a part of tei:citedRange.
            TODO: make a complete string with all non-tei:app strings and all
            tei:lem strings for highlighting. :)
            $entry/descendant::tei:lem[1]/string() => replace("[\s]+", " ")
        else
            $entry/string() => replace("[\s]+", " ")
      return
        if ($entry/@n) then
          for $reference in tokenize($entry/@n, " ")
            return
              <bdn:index-entry xmlns="http://bdn-edition.de/ns/1.0"
              id="{$chapterID}" n="{$reference}" key="{$key}" />

        else
          <bdn:index-entry xmlns="http://bdn-edition.de/ns/1.0"
          id="{$chapterID}" from="{$entry/@from}" to="{$entry/@to}" key="{$key}" />

    let $sorted-entries :=
      element bdn:root {
        index-creation:sort-bible-index-entries($extended-entries)
      }

    let $filename := index-creation:get-index-name($author, "bible-index")
    return xmldb:store($config:aux, $filename, $sorted-entries)
};


(:~
 : Sorts the Bible index entries according to the Bible.
 :
 : @author Michelle Weidling
 : @param $entries All Bible index entries of a work
 : @return a sequence that contains the entries of the Old resp. the New Tetsament
 :)
declare function index-creation:sort-bible-index-entries($entries
as element(bdn:index-entry)*) as element()*
{
  let $old-testament :=
    <bdn:ot xmlns="http://www.bdn-edition.de/ns/1.0">
      {index-creation:get-entries-in-testament($entries, "ot")}
    </bdn:ot>
  let $new-testament :=
    <bdn:nt xmlns="http://www.bdn-edition.de/ns/1.0">
      {index-creation:get-entries-in-testament($entries, "nt")}
    </bdn:nt>

  return ($old-testament, $new-testament)
};


(:~
 : Determines which Bible index entry belongs to the Old or the New Testament.
 :
 : @author Michelle Weidling
 : @param $entries All Bible index entries of a work
 : @param $testament "ot" or "nt"
 : @return a sequence that contains the entries of the Old resp. the New Testament
 :)
declare function index-creation:get-entries-in-testament($entries as element(bdn:index-entry)*,
$testament as xs:string) as element(bdn:index-entry)*
{
  let $bible := doc($config:sade-data || "2r573.xml")
  return
    for $entry in $entries
      let $entry-details :=
        if ($entry/@n) then
          tokenize($entry/@n, ':')
        else
          tokenize($entry/@from, ':')

      let $book := $entry-details[position() = 1]
      let $index-no := map:get($index-creation:bible-map, $book)
      order by $index-no,
        number($entry-details[position() = 2]),
        number($entry-details[position() = 3])
      return
        switch($testament)
          case "ot" return
            if ($index-no lt 50) then
              $entry
            else
              ()
          case "nt" return
            if ($index-no ge 50) then
              $entry
            else
              ()
          default return
            ()
};


(:~
 : A generic function for making the persons, Classics and subject index.
 :
 : @author Michelle Weidling
 : @param $author The current author
 : @param $xml The current and opened full XML file (e.g. griesbach-full.xml)
 : @param index-type The index type (e.g. "persons-index")
 : @return xs:string the location of the stored file
 :)
declare function index-creation:make-index-xml($author as xs:string,
$xml as node()*, $index-type as xs:string) as xs:string*
{
  let $entries := $xml//tei:index[@indexName = $index-type]
  let $extended-entries :=
    for $entry in $entries
      let $chapter-id := $entry/ancestor::tei:div[1]/@xml:id/string()

      let $keyword := index-creation:get-keyword($index-type, $entry)

      let $no-of-terms := count($entry//tei:term)
      return
        for $iii in 1 to $no-of-terms
          return
            <bdn:index-entry
              xmlns="http://bdn-edition.de/ns/1.0"
              key="{$keyword}"
              id="{$chapter-id}"
            >
              {$entry/tei:term[$iii]}
            </bdn:index-entry>
  let $sorted-entries :=
    for $entry in $extended-entries
      order by $entry/@key, $entry/@id
      return $entry

  let $matching-first-letters := index-creation:get-first-letters($sorted-entries)

  let $final-content :=
    element bdn:root
    {
      $matching-first-letters,
      $sorted-entries
    }

  let $filename := index-creation:get-index-name($author, $index-type)

  return xmldb:store($config:aux, $filename, $final-content)
};


(:~
 : An auxiliary function that creates a map we need to determine whether an
 : index entry is in an odd or even table row.
 :
 : @author Michelle Weidling
 : @param $author the current author
 : @param $xml an index XML that's the result of index-creation:make-(bible-)index-xml
 : @param $index-type the index type, e.g. 'subject-index'
 : @return a map indicating the position
 : :)
declare function index-creation:make-odd-even-map($author as xs:string,
$xml as node()*) as map()
{
    let $terms := $xml//tei:term
    let $distinct-terms :=
        for $term in $terms return
            if ($term/preceding::tei:term[normalize-space(./string()) = normalize-space($term/string())]) then
                ()
            else
                $term
    return
        map:merge
        (
          for $term in $distinct-terms
            let $index-no := index-of($distinct-terms, $term)
            let $odd-or-even :=
              if ($index-no mod 2 = 1) then
                "odd"
              else
                "even"
            return
              map:entry(upper-case($term/string()), $odd-or-even)
        )
};


(:~
 : Returns all characters of the alphabet that occur as first letter in a given
 : index. The result is used for displaying the alphabetical navbar on an index
 : page.
 :
 : @author Michelle Weidling
 : @param $sorted-entries the entries of a non-Bible index
 : @return an element that contains info about all occurring letters
 :)
declare function index-creation:get-first-letters($sorted-entries
as element(bdn:index-entry)*) as element(bdn:key-letters)
{
    let $keys :=
        for $entry in $sorted-entries return
            $entry/@key

    let $first-letters :=
        for $key in $keys return
            upper-case($key) => substring(1, 1)
    let $first-letters := distinct-values($first-letters)

    return
        <bdn:key-letters xmlns="http://bdn-edition.de/ns/1.0">
            {
                for $letter in $first-letters return
                  <bdn:letter id="{$letter}" />
            }
        </bdn:key-letters>
};


(:~
 : Returns the full doc name of an index.
 :
 : @author Michelle Weidling
 : @param $author the BdN author, e.g. "noesselt"
 : @param $index-type the current index, e.g. "persons-index"
 : @return the full doc name, e.g. "noesselt-persons-index.xml"
 :)
declare function index-creation:get-index-name($author as xs:string,
$index-type as xs:string)
{
  $author || "-" || $index-type || ".xml"
};


(:~
 : Returns the key of an index entry.
 :
 :)
declare function index-creation:get-keyword($index-type as xs:string,
$entry as element()) as xs:string
{
  switch ($index-type)
    case "subjects-index" return
      (: since we have some entries that start with an umlaut, these have to be
      mapped to their decomposed form in order to be sorted correctly :)
      switch(upper-case(substring($entry//tei:term,1,1)))
        case "Ä" return concat('AE', substring($entry//tei:term,2))
        case "Ö" return concat('OE', substring($entry//tei:term, 2))
        case "Ü" return concat('UE', substring($entry//tei:term, 2))
        default return
            upper-case(substring($entry//tei:term,1,1)) || substring($entry//tei:term, 2)

    case "persons-index" return
      if ($entry[count(descendant::tei:term) gt 1]) then
        (: TODO there should be an encoding for this problem by now. check and consider!:)
        replace($entry/descendant::tei:term[2]/string(), "[\s]+", " ")
      else
        replace($entry//tei:term, "[\s]+", " ")

    case "classics-index" return
      if ($entry[count(descendant::tei:persName) gt 1]) then
        replace($entry/descendant::tei:persName[2]/string(), "[\s]+", " ")
      else
        replace($entry//tei:persName, "[\s]+", " ")

    default return
      ()
};

declare function index-creation:make-all-htmls($author as xs:string,
$occuring-indices as xs:string+) as xs:string*
{
  (
    bdn-utils:assure-dir-available("indices"),
    index-creation:make-all-list-htmls($author, $occuring-indices)
  )
};

(:~
 : Creates the static list view for the differrent kinds of indices.
 :
 : @author Michelle Weidling
 : @param $author The current author
 : @return xs:string the location of the stored files
 :)
declare function index-creation:make-all-list-htmls($author as xs:string,
$occuring-indices as xs:string+) as xs:string*
{
  for $index in $occuring-indices return
    index-creation:make-list-html($author, $index)
};


(:~
 : A generic function for creating the list view for the differrent kinds of
 : indices.
 :
 : @author Michelle Weidling
 : @param $author The current author
 : @param $index-type The current kind of index, e.g. "persons-index"
 : @return xs:string the location of the stored file
 :)
declare function index-creation:make-list-html($author as xs:string,
$index-type as xs:string) as xs:string
{
  let $index-head := index-creation:make-list-index-head($index-type)
  let $index := index-creation:make-list-table($author, $index-type)
  let $index-content :=
    element xhtml:div
    {
      attribute id {$index-type},
      $index-head,
      $index
    }

  let $filename := $author || "-" || $index-type || ".html"

  return xmldb:store($config:indices, $filename, $index-content)
};


(:~
 : A generic function for creating the heading for the differrent kinds of
 : indices.
 :
 : @author Michelle Weidling
 : @param $index-type The current kind of index, e.g. "persons-index"
 : @return element(div) The heading
 :)
declare function index-creation:make-list-index-head($index-type as xs:string)
as element(div)
{
  let $title := switch($index-type)
    case "classics-index" return "Register antiker Autoren"
    case "persons-index" return "Personenregister"
    case "bible-index" return "Bibelstellenregister"
    case "subjects-index" return "Sachregister"
    default return ""

  return
    <div class="block-header">
      <h2 class="index-list-heading">
        <span class="title">{$title}</span>
        <span class="decoration" />
        <span class="decoration" />
        <span class="decoration" />
      </h2>
    </div>
};


(:~
 : A generic function for creating the content of the list view for the
 : differrent kinds of indices.
 :
 : @author Michelle Weidling
 : @param $author The current author
 : @param $index-type The current kind of index, e.g. "persons-index"
 : @return element(table) A table containing the index entries and a link to
 : them in the respective files
 :)
declare function index-creation:make-list-table($author as xs:string,
$index-type as xs:string) as element(table)
{
  let $xml := doc($config:aux || $author || "-" || $index-type || ".xml")
  let $entries := $xml//bdn:index-entry
  let $odd-even-map := index-creation:make-odd-even-map($author, $xml)

  return
    <table class="{concat("index-table ", $index-type)}">
      {
        (: for the bible index we create two table headings: one for the
        old and one for the new testament :)
        if ($index-type = "bible-index") then
          (
            index-creation:make-table-head($index-type || "-ot"),
            for $entry in $xml//bdn:ot/bdn:index-entry return
              index-creation:make-bible-table-entry($entry, $index-type),
              index-creation:make-table-head($index-type || "-nt"),
            for $entry in $xml//bdn:nt/bdn:index-entry return
              index-creation:make-bible-table-entry($entry, $index-type)
          )
        else
          (index-creation:make-table-head($index-type),
          for $entry in $entries return
            index-creation:make-table-entry($entry, $index-type, $odd-even-map))
      }
    </table>
};

declare function index-creation:make-table-head($index-type as xs:string)
as element(tr)*
{
  let $label :=
    switch($index-type)
      case "bible-index-ot" return "Bibelstelle"
      case "bible-index-nt" return "Bibelstelle"
      case "subjects-index" return "Begriff"
      default return "Name"
  let $tr-class := "index-table-head " || $index-type || "-label-td"

  return
    (
      if (matches($index-type, "bible")) then
        <tr class ="index-table-head-bible" id="{$index-type}">
          <td class="{concat($index-type, "-label-td")}">
            {
              if ($index-type = "bible-index-ot") then
                "Altes Testament"
              else
                "Neues Testament"
            }
          </td>
          <td />
        </tr>
      else
        (),
      <tr class ="index-table-head">
        <td class="{concat($index-type, "-label-td")}">
          <strong>{$label} </strong>
          <!--<a href="{$url || $register || "&#38;mode=default"}">
          </a>-->
        </td>
        {
          if ($index-type = "classics-index") then
            <td class="classics-index-work-td">
              <strong>Werk </strong>
            </td>
          else
            ()
        }
        <td class="{concat($index-type, "-source-td")}">
          <strong>Nachweis </strong>
          <!--<a href="{$url || $register || "&#38;mode=chapter"}">
          </a>-->
        </td>
      </tr>
    )
};

declare function index-creation:make-table-entry($entry as element(bdn:index-entry),
$index-type as xs:string, $odd-even-map as map()) as element(tr)*
{
  let $first-letter := substring($entry/@key, 1, 1)
  let $tag :=
    if ($entry/preceding-sibling::*[starts-with(@key, $first-letter)]) then
      ()
    else
      index-creation:make-tag($first-letter, $index-type)

  let $term-already-mentioned := index-creation:has-term-already-been-mentioned($entry)

  let $table-key :=
    if ($term-already-mentioned
    (: ", s." denote crossreferences which shouldn't be displayed :)
    or matches($entry/@key, ", s.")) then
      ()
    else
      string($entry/@key)

    let $odd-or-even := map:get($odd-even-map, upper-case($entry/@key))

  let $class :=
    if ($term-already-mentioned) then
      "already-mentioned " || $odd-or-even
    else
      "index-entry " || $odd-or-even

  let $target-url := index-creation:make-url($entry/@id, $entry/@key)

  return
    (
      $tag,
      <tr class="{$class}">
        <td>{$table-key}</td>
        <td>
          <a href="{$target-url}" target="_blank">
            <i class="fa fa-link" />
          </a>
        </td>
      </tr>
    )
};

declare function index-creation:make-bible-table-entry($entry as element(bdn:index-entry),
$index-type as xs:string) as element(tr)*
{
  let $bibl-already-mentioned := index-creation:has-bibl-already-been-mentioned($entry)
  let $bible-reference :=
    if ($bibl-already-mentioned) then
      ()
    else
      index-creation:make-bible-reference($entry)

  let $class :=
    if ($bibl-already-mentioned) then
      "already-mentioned "
    else
      "index-entry "

  let $target-url := index-creation:make-url($entry/@id, $entry/@key)

  return
    <tr class="{$class}">
      <td>{$bible-reference}</td>
      <td>
        <a href="{$target-url}" target="_blank">
          <i class="fa fa-link" />
        </a>
      </td>
    </tr>
};

declare function index-creation:has-term-already-been-mentioned($entry
as element(bdn:index-entry)) as xs:boolean
{
  if ($entry/preceding::*[@key = $entry/@key]) then
    true()
  else
    false()
};

declare function index-creation:has-bibl-already-been-mentioned($entry
as element(bdn:index-entry)) as xs:boolean
{
  if ($entry/preceding::*[@n = $entry/@n]
  or ($entry/preceding::*[@from = $entry/@from]
    and $entry/preceding::*[@to = $entry/@to])) then
    true()
  else
    false()
};

(:~
 : Creates the table entry for a Bible reference. A Bible reference has the
 : pattern $book:$chapter:$verse while $verse is optional.
 :
 : The resulting string should contain as little duplication of information as
 : possible, e.g. @from="Gen:1:1" @to="Gen:1:19" is rendered to Gen 1,1-19 since
 : both verse 1 and verse 19 refer to the same book within Genesis.
 :
 : The creation of Bible references is heavily borrowed by the algorithm we use
 : for the Bible index in the printed edition.
 :
 : It still lacks a solution for ranges that span over different biblical books,
 : e.g. @from="Gen:1:1" @to="Ex:2:5" but it is unclear if this is ever the
 : case.
 :
 : @author Michelle Weidling
 : @param $entry The current entry in the Bible index
 : @return A string with the full Bible reference
 :)
declare function index-creation:make-bible-reference($entry as element(bdn:index-entry))
as xs:string
{
  if ($entry/@n) then
    let $entry-details := tokenize($entry/@n, ":")
    let $book := $entry-details[1]
    let $chapter := $entry-details[2]
    let $verse := $entry-details[3]
    let $passage := concat($chapter, ',', $verse)

    return
      if ($verse) then
        $book || " " || $passage
      else
        $book || " " || $chapter
  else
    let $from := tokenize($entry/@from, ":")
    let $from-book := $from[1]
    let $from-chapter := $from[2]
    let $from-verse := $from[3]
    let $from-passage := concat($from-chapter, ',', $from-verse)
    let $to := tokenize($entry/@to, ":")
    let $to-book := $to[1]
    let $to-chapter := $to[2]
    let $to-verse := $to[3]
    let $to-passage := concat($to-chapter, ',', $to-verse)

    let $from-statement :=
      if ($from-verse) then
        $from-book || " " || $from-passage
      else
        $from-book || " " || $from-chapter

    let $to-statement :=
      if (matches($to-book, "f")) then
        $to-book || "."
      else if ($from-book = $to-book
        and $from-chapter = $to-chapter) then
        text{"-" || $to-verse}
      else if ($from-book = $to-book) then
        if ($to-verse) then
          $to-passage
        else
          $to-chapter
      else
        ()
    return $from-statement || $to-statement
};

declare function index-creation:make-url($id as xs:string, $keyword as xs:string)
as xs:string
{
  let $file-uri := map:get($index-creation:id-file-map, $id)
  return "book.html?file=" || $file-uri || "&amp;edition=app&amp;highlight=" || $keyword || "#" || $id
};

(:~
 : Determines which index types occur in a work. The occurring index types are
 : added to the whitelist.
 :
 : @author Michelle Weidling
 : @param $xml the current full XML
 : @param $author the current author
 :)
declare function index-creation:determine-occurring-indices($xml as node()*,
$author as xs:string) as map(*)
{
  let $index-map :=
    map:merge
    (
      for $type in $index-creation:index-types return
        map:entry
        (
          $type,
          switch ($type)
            case "bible-index" return
              if ($xml//tei:bibl[@type = "biblical-reference"]/tei:citedRange) then
                true()
              else
                false()
            default return
              if ($xml//tei:index[@indexName = $type]) then
                true()
              else
                false()
        )
    )

  let $add-to-whitelist := whitelist:add-index-type($index-map, $author)

  return
      $index-map
};

(:~
 : Creates a tag for a first letters of the alphabet occurring in an index type.
 :
 :)
declare function index-creation:make-tag($first-letter as xs:string,
$index-type as xs:string) as element(tr)*
{
  let $tag-is-a := boolean($first-letter = "A")
  let $css-style :=
    if ($tag-is-a) then
      'font-size: x-large'
    else
      'padding-top: 90px; font-size: x-large'

  return
    <tr class="tag">
      <td id="{$first-letter}" class="{if ($tag-is-a) then "tag-a" else "tag-non-a"}">
        <b>{$first-letter}</b>
      </td>
      <td />
        {if ($index-type = "classics-index") then <td /> else ()}
    </tr>
};
